/'****************************************************************************
*
* Name: tableobj.bi
*
* Synopsis: Table object source file.
*
* Description: The table object provides a list object that can store different
*              type of data within each row of the object, much like a listview
*              or database.
*
* Copyright 2009, Richard D. Clark
*
*                          The Wide Open License (WOL)
*
* Permission to use, copy, modify, distribute and sell this software and its
* documentation for any purpose is hereby granted without fee, provided that
* the above copyright notice and this license appear in all source copies. 
* THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY OF
* ANY KIND. See http://www.dspguru.com/wol.htm for more information.
*
*****************************************************************************'/

'Object namespace.
Namespace tableobj

Const tblVersion = "0.1"

'Create a NULL value.
#Ifndef NULL
    #Define NULL 0
#EndIf

'Create a True/False value.
#Ifndef FALSE
    #Define FALSE 0
    #Define TRUE (Not FALSE)
#EndIf

'Enumeration of error codes.
Enum ecodes
    etNoError
    etNoMemory
    etBadRowParameter
    etBadColParameter
End Enum

'Enumeration of allowable types in each cell.
Enum cdatatype
    tNull
    tByte
    tUByte
    tShort
    tUShort
    tInteger
    tUInteger
    tLong
    tUlong
    tLongint
    tUlongint
    tSingle
    tDouble
    tString
End Enum

'Structure used in the table object to represent each cell.
Type tcell
    Private:
    _ctype As cdatatype 'Current type of data, default is null.
    Union
        _clbyte As Byte     'Depending on the value of ctype, cl* will contain the data.
        _clubyte As UByte    'If ctype is tNull, the cell will not contain any data.
        _clshort As Short
        _clushort As UShort
        _clinteger As Integer
        _cluinteger As UInteger
        _cllong As Long
        _clulong As ULong
        _cllongint As LongInt
        _clulongint As ULongInt
        _clsingle As Single
        _cldouble As Double
        _clstring As ZString Ptr
    End Union
    Declare Sub _ClearCell () 
    Public:
    Declare Constructor ()
    Declare Destructor ()
    Declare Property DataType () As cdatatype 'Returns the cell data type.
    Declare Sub ClearCell ()
    Declare Operator Let (rhs As tcell) 'Sets data using tcell object.
    Declare Operator Let (rhs As Byte) 'Sets object data using variable.
    Declare Operator Let (rhs As UByte) 
    Declare Operator Let (rhs As Short) 
    Declare Operator Let (rhs As UShort) 
    Declare Operator Let (rhs As Integer) 
    Declare Operator Let (rhs As UInteger) 
    Declare Operator Let (rhs As Long) 
    Declare Operator Let (rhs As ULong) 
    Declare Operator Let (rhs As LongInt)
    Declare Operator Let (rhs As ULongInt)
    Declare Operator Let (rhs As Single)
    Declare Operator Let (rhs As Double)
    Declare Operator Let (rhs As String)
    Declare Operator Cast () As Byte  'Cast to data type.
    Declare Operator Cast () As UByte
    Declare Operator Cast () As Short
    Declare Operator Cast () As UShort
    Declare Operator Cast () As Integer
    Declare Operator Cast () As UInteger
    Declare Operator Cast () As Long
    Declare Operator Cast () As ULong
    Declare Operator Cast () As LongInt
    Declare Operator Cast () As ULongInt
    Declare Operator Cast () As Single
    Declare Operator Cast () As Double
    Declare Operator Cast () As String
End Type

'Clears nay data in cell and sets to Null.
Sub tcell._ClearCell()

    'Based on what the type is, clear the data.
    Select Case _ctype
        Case tNull
            'Nothing to do here.
        Case tByte
            _clbyte = 0
        Case tUByte
            _clubyte = 0
        Case tShort
            _clshort = 0
        Case tUShort
            _clushort = 0
        Case tInteger
            _clinteger = 0
        Case tUInteger
            _cluinteger = 0
        Case tLong
            _cllong = 0
        Case tUlong
            _clulong = 0
        Case tLongint
            _cllongint = 0
        Case tUlongint
            _cllongint = 0
        Case tSingle
            _clsingle = 0
        Case tDouble
            _cldouble = 0
        Case tString
            'Deallocate string if present and set ptr to null.
            If _clstring <> NULL Then
                DeAllocate _clstring
                _clstring = NULL
            EndIf
    End Select
    'Set data type to tNull.
    _ctype = tNull
End Sub

'Cell constructor sets cell to Null.
Constructor tcell ()
    _ctype = tNull
End Constructor

'Calls _ClearCell to clear contents.
Destructor tcell ()
    _ClearCell
End Destructor

'Calls _ClearCell to clear contents.
Sub tcell.ClearCell ()
    _ClearCell
End Sub

'Returns the cell data type.
Property tcell.DataType () As cdatatype
    Return _ctype
End Property

Operator tcell.Let (rhs As tcell)
    
    _ClearCell
    'Based on what the type is, clear the data.
    _ctype = rhs.DataType
    Select Case _ctype
        Case tNull
            'Nothing to do here.
        Case tByte
            _clbyte = rhs
        Case tUByte
            _clubyte = rhs
        Case tShort
            _clshort = rhs
        Case tUShort
            _clushort = rhs
        Case tInteger
            _clinteger = rhs
        Case tUInteger
            _cluinteger = rhs
        Case tLong
            _cllong = rhs
        Case tUlong
            _clulong = rhs
        Case tLongint
            _cllongint = rhs
        Case tUlongint
            _clulongint = rhs
        Case tSingle
            _clsingle = rhs
        Case tDouble
            _cldouble = rhs
        Case tString
            Dim As String s = rhs
            Dim As Integer l = Len(s)
            If l > 0 Then
                _clstring = Allocate(l + 1)
                *_clstring = s
            Else
                _clstring = NULL
            EndIf
            
    End Select
    
End Operator

'Sets object data using variable.
Operator tcell.Let (rhs As Byte) 
    
    'Clear any existing data.
    _ClearCell
    _ctype = tByte
    _clbyte = rhs
    
End Operator

Operator tcell.Let (rhs As UByte)
    
    'Clear any existing data.
    _ClearCell
    _ctype = tUByte
    _clubyte = rhs

End Operator

Operator tcell.Let (rhs As Short)
    
    'Clear any existing data.
    _ClearCell
    _ctype = tShort
    _clshort = rhs

End Operator

Operator tcell.Let (rhs As UShort)
    
    'Clear any existing data.
    _ClearCell
    _ctype = tUShort
    _clushort = rhs

End Operator

Operator tcell.Let (rhs As Integer)
    
    'Clear any existing data.
    _ClearCell
    _ctype = tInteger
    _clinteger = rhs

End Operator

Operator tcell.Let (rhs As UInteger)
    
    'Clear any existing data.
    _ClearCell
    _ctype = tUinteger  
    _cluinteger = rhs
End Operator

Operator tcell.Let (rhs As Long)
    
    'Clear any existing data.
    _ClearCell
    _ctype = tLong
    _cllong = rhs

End Operator

Operator tcell.Let (rhs As ULong)
    
    'Clear any existing data.
    _ClearCell
    _ctype = tULong
    _clulong = rhs

End Operator

Operator tcell.Let (rhs As LongInt)
    
    'Clear any existing data.
    _ClearCell
    _ctype = tLongint
    _cllongint = rhs

End Operator

Operator tcell.Let (rhs As ULongInt)
    
    'Clear any existing data.
    _ClearCell
    _ctype = tULongint
    _clulongint = rhs

End Operator

Operator tcell.Let (rhs As Single)
    
    'Clear any existing data.
    _ClearCell
    _ctype = tSingle
    _clsingle = rhs

End Operator

Operator tcell.Let (rhs As Double)
    
    'Clear any existing data.
    _ClearCell
    _ctype = tDouble
    _cldouble = rhs
    
End Operator

Operator tcell.Let (rhs As String)
    Dim As Integer l = Len(rhs)
    
    'Clear any existing data.
    _ClearCell
    _ctype = tString
    If l > 0 Then
        _clstring = Allocate(l + 1)
        *_clstring = rhs
    Else
        _clstring = NULL
    EndIf
    
End Operator
'Cast to data type.
Operator tcell.Cast () As Byte
    If _ctype = tByte Then
        Return _clbyte
    EndIf
End Operator

Operator tcell.Cast () As UByte
    If _ctype = tuByte Then
        Return _clubyte
    EndIf
End Operator

Operator tcell.Cast () As Short
    If _ctype = tShort Then
        Return _clshort
    EndIf
End Operator

Operator tcell.Cast () As UShort
    If _ctype = tUShort Then
        Return _clushort
    EndIf
End Operator

Operator tcell.Cast () As Integer
    If _ctype = tInteger Then
        Return _clinteger
    EndIf
End Operator

Operator tcell.Cast () As UInteger
    If _ctype = tUInteger Then
        Return _cluinteger
    EndIf
End Operator

Operator tcell.Cast () As Long
    If _ctype = tLong Then
        Return _cllong
    EndIf
End Operator

Operator tcell.Cast () As ULong
    If _ctype = tULong Then
        Return _clulong
    EndIf
End Operator

Operator tcell.Cast () As LongInt
    If _ctype = tLongInt Then
        Return _cllongint
    EndIf
End Operator

Operator tcell.Cast () As ULongInt
    If _ctype = tUlongInt Then
        Return _clulongint
    EndIf
End Operator

Operator tcell.Cast () As Single
    If _ctype = tSingle Then
        Return _clsingle
    EndIf
End Operator

Operator tcell.Cast () As Double
    If _ctype = tDouble Then
        Return _cldouble
    EndIf
End Operator

Operator tcell.Cast () As String
    If _ctype = tString Then
        If _clstring <> NULL Then
            Return *_clstring
        Else
            Return ""
        EndIf
    EndIf
End Operator


'Table object.
Type tblobj
    Private:
    _tdata As tcell Ptr Ptr  'The table of cells. This is a set of rows with variable columns.
    _trowcnt As Integer  'Number of rows in table.
    _tcolcnt As Integer Ptr 'The number of columns in each row.
    _version As String  'Current version of table object.
    _errcode As ecodes  'Current error if any.
    Declare Sub _DestroyTable () 'Releases all allocated memory of the table and resets object.
    Declare Function _AddRow () As Integer 'Adds a new row to the table.
    Declare Function _AddCol (row As Integer) As Integer 'Adds a new bolumnd at row to the table.
    Public:
    Declare Constructor () 'Creates an empty table.
    Declare Destructor () 'Calls _DestroyTable.
    Declare Property Version () As String 'Returns the current version.
    Declare Property ErrorCode () As ecodes 'Returns current error code as clears error flag.
    Declare Property RowCount () As Integer 'Returns the current number of rows.
    Declare Function ColCount (row As Integer) As Integer 'Returns the current number of columns in row.
    Declare Function ErrorString (errcode As ecodes) As String 'Returns error string.
    Declare Function AddRow () As Integer 'Adds a row to table and returns row index.
    Declare Function AddColumn (row As Integer, cdata As tcell) As Integer 'Adds column to row.
    Declare Function ColDataType (row As Integer, col As Integer) As cdatatype 'Returns the current data type of column in row.
    Declare Sub ClearCol (row As Integer, col As Integer) 'Sets column to null.
    Declare Sub SetColData (row As Integer, col As Integer, cdata As tcell) 'Sets the column data. 
    Declare Sub GetColData (row As Integer, col As Integer, ByRef cdata As tcell)  'Gets column in row as byte. 
    Declare Sub DestroyTable () 'Calls _DestroyTable.
End Type

'Calls _DestroyTable to deallocate table.
Sub tblobj._DestroyTable ()
    'Make sure we have some items in the table.
    If _tdata <> NULL Then
        'Delete each row of the table.
        For row As Integer = _trowcnt - 1 To 0 Step -1
            'Clear each cell in row.
            For col As Integer = _tcolcnt[row] - 1 To 0 Step -1
                _tdata[row][col].ClearCell
            Next 
            DeAllocate _tdata[row]
        Next
        'Delete the table.
        DeAllocate _tdata
        'Delete the column count.
        DeAllocate _tcolcnt
    EndIf
    'Set the table ptr to null.
    _tdata = NULL
    'Set the column count to null.
    _tcolcnt = NULL
    'Reset the list count.
    _trowcnt = 0
End Sub

'Adds a new row to the table. Returns TRUE if successful.
Function tblobj._AddRow () As Integer
    Dim As Integer ret = TRUE, rowcnt = _trowcnt
    
    'Increment the row count.
    rowcnt += 1
    'Try and allocate new row.
    _tdata = ReAllocate(_tdata, rowcnt * SizeOf(tcell Ptr))
    If _tdata <> NULL Then
        'Set new row cnt.
        _trowcnt = rowcnt
        'Set the new position to NULL.
        _tdata[_trowcnt - 1] = NULL
        'Add new column cnt.
        _tcolcnt = ReAllocate(_tcolcnt, _trowcnt * SizeOf(Integer))
        'Make sure we could allocate memory.
        If _tcolcnt <> NULL Then
            'Set count to 0.
            _tcolcnt[_trowcnt - 1] = 0
        Else
            ret = FALSE
        EndIf
    Else
        ret = FALSE
    EndIf
    
    Return ret
End Function

'Adds a new column at row.
Function tblobj._AddCol (row As Integer) As Integer
    Dim As Integer ret = TRUE, colcnt = _tcolcnt[row]
    
    'Note: Don't need to check row here as it is validated before this method is called.
    'Increment the column count at row.
    colcnt += 1
    'Add new column at row.
    _tdata[row] = ReAllocate(_tdata[row], colcnt * SizeOf(tcell))
    'Make sure column was allocated. 
    If _tdata[row] <> NULL Then
        'Set the new column count.
        _tcolcnt[row] = colcnt
    Else
        'Memory error.
        ret = FALSE
    EndIf
    Return ret
End Function

'Creates an empty table.
Constructor tblobj ()
    'Clear current list if any.
    _DestroyTable
    'Set current internal variables.
    _version = tblVersion
    _errcode = etNoError
End Constructor

'Object destructor.
Destructor tblobj ()
    _DestroyTable
End Destructor

'Returns the current version.
Property tblobj.Version () As String
	Return _version
End Property

'Returns current error code and clears error flag.
Property tblobj.ErrorCode () As ecodes
    Dim tmp As ecodes
    
    tmp = _errcode
    _errcode = etNoError
    
    Return tmp
End Property

'Returns the current number of rows.
Property tblobj.RowCount () As Integer 
    Return _trowcnt
End Property

'Returns error string of error code.
Function tblobj.ErrorString (errcode As ecodes) As String
    Select Case errcode
        Case etNoError: Return "No error."
        Case etNoMemory: Return "Could not allocate memory for object."
        Case etBadRowParameter: Return "Bad row parameter."
        Case etBadColParameter: Return "Bad column parameter."
        Case Else
            Return "Unknown error code."
    End Select
End Function

'Returns the current number of columns in row.
Function tblobj.ColCount (row As Integer) As Integer
    Dim As Integer ret = -1
    
    'Set default error.
    _errcode = etNoError
    If _tdata <> NULL Then
        'Make sure row is valid.
        If (row >= 0) And (row <= _trowcnt - 1) Then
            'Get the number of columns in row.
            ret = _tcolcnt[row]
        Else
            'Bad row parameter.
            _errcode = etBadRowParameter    
        EndIf
    Else
        _errcode = etBadRowParameter
    End If
    
    Return ret
End Function

'Adds a row to table and returns row index.
Function tblobj.AddRow () As Integer 
    Dim As Integer ret, index = -1
    
    'Set default error code.
    _errcode = etNoError
    ret = _AddRow ()
    'Could allocate new row.
    If ret = FALSE Then
        _errcode = etNoMemory
    Else
        index = _trowcnt - 1 
    EndIf
    
    Return index
End Function

'Adds column to row and returns column index.
Function tblobj.AddColumn (row As Integer, cdata As tcell) As Integer 
    Dim As Integer ret, index = -1

    If _tdata <> NULL Then
        'Set default error.
        _errcode = etNoError
        'Make sure row is valid.
        If (row >= 0) And (row <= _trowcnt - 1) Then
            'Add new cell.
            ret = _AddCol (row)
            'make sure cell was allocated.
            If ret = FALSE Then
                _errcode = etNoMemory
            Else
                'Get the index value.
                index = _tcolcnt[row] - 1
                'Set the data type and data.
                 _tdata[row][index] = cdata
            EndIf
        Else
            'Bad row parameter.
            _errcode = etBadRowParameter    
        EndIf
    Else
        _errcode = etBadRowParameter    
    End If
    
    Return index
End Function

Function tblobj.ColDataType (row As Integer, col As Integer) As cdatatype
    Dim As cdatatype ret = tNull
    
    If _tdata <> NULL Then
        'Set default error.
        _errcode = etNoError
        'Make sure row and col is valid.
        If (row >= 0) And (row <= _trowcnt - 1) Then
            If _tdata[row] <> NULL Then
                If (col >= 0) And (col <= _tcolcnt[row] - 1) Then
                    ret = _tdata[row][col].DataType
                Else
                    'Bad col parameter.
                    _errcode = etBadColParameter    
                End If
            Else
                _errcode = etBadColParameter    
            End If
        Else
            'Bad row parameter.
            _errcode = etBadRowParameter    
        EndIf
    Else
        _errcode = etBadRowParameter    
    End If
    
    Return ret  
End Function


'Clears any data and set column to tNull.
Sub tblobj.ClearCol (row As Integer, col As Integer)

    If _tdata <> NULL Then
        'Set default error.
        _errcode = etNoError
        'Make sure row and col is valid.
        If (row >= 0) And (row <= _trowcnt - 1) Then
            If _tdata[row] <> NULL Then
                If (col >= 0) And (col <= _tcolcnt[row] - 1) Then
                    'Clear current data.
                    _tdata[row][col].ClearCell 
                Else
                    'Bad col parameter.
                    _errcode = etBadColParameter    
                End If
            Else
                _errcode = etBadColParameter    
            End If
        Else
            'Bad row parameter.
            _errcode = etBadRowParameter    
        EndIf
    Else
        _errcode = etBadRowParameter    
    End If
    
End Sub

'Sets column in row to byte.
Sub tblobj.SetColData (row As Integer, col As Integer, cdata As tcell)

    If _tdata <> NULL Then
        'Set default error.
        _errcode = etNoError
        'Make sure row and col is valid.
        If (row >= 0) And (row <= _trowcnt - 1) Then
            If _tdata[row] <> NULL Then
                If (col >= 0) And (col <= _tcolcnt[row] - 1) Then
                    'Clear current data.
                    _tdata[row][col].ClearCell
                    'Set data to passed data type.
                    _tdata[row][col] = cdata
                Else
                    'Bad col parameter.
                    _errcode = etBadColParameter    
                End If
            Else
                _errcode = etBadColParameter    
            End If
        Else
            'Bad row parameter.
            _errcode = etBadRowParameter    
        EndIf
    Else
        _errcode = etBadRowParameter    
    End If
    
End Sub

'Gets column in row as byte.
Sub tblobj.GetColData (row As Integer, col As Integer, ByRef cdata As tcell)
    
    If _tdata <> NULL Then
        'Set default error.
        _errcode = etNoError
        'Make sure row and col is valid.
        If (row >= 0) And (row <= _trowcnt - 1) Then
            If _tdata[row] <> NULL Then
                If (col >= 0) And (col <= _tcolcnt[row] - 1) Then
                    'Clear any data in cdata.
                    cdata.ClearCell
                    cdata = _tdata[row][col]
                Else
                    'Bad col parameter.
                    _errcode = etBadColParameter    
                End If
            Else
                _errcode = etBadColParameter    
            End If
        Else
            'Bad row parameter.
            _errcode = etBadRowParameter    
        EndIf
    Else
        _errcode = etBadRowParameter    
    End If
    
End Sub

'Deletes table and resets object.
Sub tblobj.DestroyTable ()
    _DestroyTable
End Sub

End Namespace
